# Standard imports
import io
import os
import argparse
import json
from pprint import pprint

# Third-party imports
import requests
import pandas as pd

# Project imports
from .settings import ENV


PARAMS_COERCION = {  # Convert these names in the params file
    "test_train_split": "train_test_split",  # Common mistake
    "num_training_examples": "train_test_split",  #  TODO: Think of something better
    "dataset": "filename",
}
TRAIN_CAMPAIGN_DEV_URL = "https://pdersvjxmgrkqojwyeyocqm7le0iwtkx.lambda-url.eu-west-2.on.aws/"
TRAIN_CAMPAIGN_STAGE_URL = "https://b7vgjlsn73e7n7kci5rwoxjc7e0pkcqn.lambda-url.eu-west-2.on.aws/"
TRAIN_CAMPAIGN_CLOUD_URL = "https://cma567qwquixu7bbjcnhbjsxjm0yumvu.lambda-url.eu-west-2.on.aws/"


### Utility functions ###


# def unwrap_payload(event: dict) -> dict:
#     """
#     Return payload and decode if it is base64 encoded
#     TODO: Not used yet...
#     """
#     if "body" not in event:  # Get body
#         raise Exception("No body in request")
#     body = event["body"]
#     if "isBase64Encoded" in event:  # Decode
#         if event["isBase64Encoded"]:
#             body = base64.b64decode(body)
#     try:  # Parse
#         payload = json.loads(body)
#     except:
#         raise Exception("Could not parse body as JSON")
#     return payload


def get_command_line_args() -> argparse.Namespace:
    """
    Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "server",
        default="cloud",
        nargs='?',
        help="Specify whether to use local or cloud lambda function",
    )
    args = parser.parse_args()
    return args


def construct_standard_headers(debug=False) -> dict:
    headers = {
        "X-Group": ENV.TWINLAB_GROUPNAME,
        "X-User": ENV.TWINLAB_USERNAME,
        "authorizationToken": ENV.TWINLAB_TOKEN,
        "X-Debug": str(debug).lower(),
    }
    return headers


def get_server_url(server: str) -> str:
    """
    The URL is the dockerised lambda function that's been set up in cloud by alexander
    """
    if server == "local":
        baseURL = ENV.TWINLAB_LOCAL_SERVER
    elif server == "dev":
        baseURL = ENV.TWINLAB_DEV_SERVER
    elif server == "stage":
        baseURL = ENV.TWINLAB_STAGE_SERVER
    elif server == "cloud":
        baseURL = ENV.TWINLAB_SERVER
    else:
        print("Server:", server)
        raise ValueError(
            "Server must be either 'local', 'dev', 'stage', or 'cloud'")
    if baseURL is None:  # Catch if the server URL has not been set
        raise ValueError("Server URL not set")
    return baseURL


def get_train_campaign_url(server: str) -> str:
    """
    Get the URL for the train_campaign lambda function
    These are different to avoid the AWS Lambda 29s gateway timeout
    """
    if server == "dev":
        url = TRAIN_CAMPAIGN_DEV_URL
    elif server == "stage":
        url = TRAIN_CAMPAIGN_STAGE_URL
    elif server == "cloud":
        url = TRAIN_CAMPAIGN_CLOUD_URL
    else:
        url = os.path.join(get_server_url(server), "train_campaign")
    return url


def coerce_params_dict(params: dict) -> dict:
    """
    Relabel parameters to be consistent with twinLab library
    """
    for param in PARAMS_COERCION:
        if param in params:
            params[PARAMS_COERCION[param]] = params.pop(param)
    return params


### ###

### HTTP requests ###


def upload_file_to_presigned_url(file_path: str, url: str, verbose=False) -> None:
    """
    Upload a file to the specified pre-signed URL.
    params:
        file_path: str; the path to the local file you want to upload.
        presigned_url: The pre-signed URL generated for uploading the file.
        verbose: bool
    """

    with open(file_path, "rb") as file:
        headers = {"Content-Type": "application/octet-stream"}
        response = requests.put(url, data=file, headers=headers)
    if verbose:
        if response.status_code == 200:
            print(f"File {file_path} uploaded successfully.")
        else:
            print(f"File upload failed")
            print(f"Status code: {response.status_code}")
            print(f"Reason: {response.text}")
        print()


def upload_dataframe_to_presigned_url(dataset_name: str, df: pd.DataFrame, url: str, verbose=False) -> None:
    """
    Upload a panads dataframe to the specified pre-signed URL.
    params:
        df: The pandas dataframe to upload
        presigned_url: The pre-signed URL generated for uploading the file.
    """
    headers = {"Content-Type": "application/octet-stream"}
    buffer = io.BytesIO()
    df.to_csv(buffer, index=False)
    buffer = buffer.getvalue()
    response = requests.put(url, data=buffer, headers=headers)
    if verbose:
        if response.status_code == 200:
            print(f"Dataframe uploaded successfully.")
        else:
            print(f"File upload failed")
            print(f"Status code: {response.status_code}")
            print(f"Reason: {response.text}")
        print()


def extract_csv_from_response(response: requests.Response, name: str) -> pd.DataFrame:
    """
    Extract CSV from response
    """
    body = response.json()  # Get the body of the response as a dictionary
    data = body[name]  # Get the entry corresponding to the field name
    df = pd.read_json(data, orient="split")
    return df


def extract_item_from_response(response: requests.Response, name: str) -> any:
    """
    Extract CSV from response
    """
    body = response.json()  # Get the body of the response as a dictionary
    item = body[name]  # Get the entry corresponding to the field name
    return item


def print_response_headers(r: requests.Response) -> None:
    """
    Print response headers
    """
    print("Response headers:")
    pprint(dict(r.headers))
    print()


def print_response_message(r: requests.Response) -> None:
    """
    Print response message
    """
    print("Response:", json.loads(r.text)["message"])
    print()


def check_response(r: requests.Response) -> None:
    if r.status_code != 200:
        print("Status code:", r.status_code)
        print_response_message(r)
        raise RuntimeError("Response error")

### ###



# Standard imports
from typing import Optional

# Third-party imports
from pydantic import BaseSettings


class Environment(BaseSettings):
    TWINLAB_LOCAL_SERVER: Optional[str]
    TWINLAB_DEV_SERVER: Optional[str]
    TWINLAB_STAGE_SERVER: Optional[str]
    TWINLAB_SERVER: str
    TWINLAB_GROUPNAME: str
    TWINLAB_USERNAME: str
    TWINLAB_TOKEN: str

    class Config:
        env_prefix = ""
        case_sensitive = False
        env_file = ".env"
        env_file_encoding = "utf-8"


ENV = Environment()

print(
    f"""
      == TwinLab Client Initialisation ==
      Server  : {ENV.TWINLAB_SERVER}
      Group   : {ENV.TWINLAB_GROUPNAME}
      User    : {ENV.TWINLAB_USERNAME}
      """
)



# Standard imports
import io
import json
from pprint import pprint
from typing import Union

# Third-party imports
import requests
import pandas as pd

# Project imports
from . import utils

### Dataset functions ###


def upload_dataset(filepath_or_df: Union[str, pd.DataFrame], dataset_name=None, server="cloud", verbose=False, debug=False) -> None:
    """
    Upload a dataset to the cloud so that it can be queried and used for training
    params:
        filepath_or_df: str; location of csv dataset on local machine or pandas dataframe
        dataset_name: str; name of dataset on S3 (same as the uploaded file name if file)
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """

    # Sort out dataset_name
    if dataset_name is None:
        if type(filepath_or_df) == str:
            dataset_name = filepath_or_df
        else:
            raise ValueError("Name must be specified if uploading dataframe")
    else:
        if "/" in dataset_name:
            raise ValueError("Dataset name cannot contain '/'")

    # Get the upload URL
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Dataset"] = dataset_name
    lambda_url = utils.get_server_url(server) + "/generate_upload_url"
    r = requests.get(lambda_url, headers=headers)
    utils.check_response(r)

    # Upload the file
    if verbose:
        utils.print_response_message(r)
    upload_url = r.json()["url"]
    if type(filepath_or_df) is str:
        filepath = filepath_or_df
        utils.upload_file_to_presigned_url(
            filepath, upload_url, verbose=verbose)
    else:
        df = filepath_or_df
        utils.upload_dataframe_to_presigned_url(dataset_name,
                                                df, upload_url, verbose=verbose)
    if verbose:
        print(f"Uploading {dataset_name}")

    # Process the uploaded dataset remotely
    process_url = utils.get_server_url(server) + "/process_uploaded_dataset"
    r = requests.post(process_url, headers=headers)
    if verbose:
        utils.print_response_message(r)


def query_dataset(dataset_name: str, server="cloud", verbose=False, debug=False) -> pd.DataFrame:
    """
    Query a dataset that exists on the cloud by printing summary statistics
    params:
        dataset_name: str; name of dataset on S3 (same as the uploaded file name)
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/query_dataset"
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Dataset"] = dataset_name
    r = requests.get(url, headers=headers)
    utils.check_response(r)
    df = utils.extract_csv_from_response(r, "summary")
    if verbose:
        utils.print_response_message(r)
        print("Summary:\n", df, "\n")
    return df


def list_datasets(server="cloud", verbose=False, debug=False) -> Union[list, None]:
    """
    List datasets that have been uploaded to the cloud
    params:
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/list_datasets"
    headers = utils.construct_standard_headers(debug=debug)
    r = requests.get(url, headers=headers)
    utils.check_response(r)
    if verbose:
        utils.print_response_message(r)
    datasets = r.json()["datasets"]
    if verbose and datasets:
        print("Datasets:")
        pprint(datasets)
    return datasets


def delete_dataset(dataset_name: str, server="cloud", verbose=False, debug=False) -> None:
    """
    Delete a dataset from the cloud
    params:
        dataset_name: str; name of dataset on S3 (same as the uploaded file name)
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/delete_dataset"
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Dataset"] = dataset_name
    r = requests.post(url, headers=headers)
    utils.check_response(r)
    if verbose:
        utils.print_response_message(r)

###  ###

### Campaign functions ###


def train_campaign(filepath_or_params: Union[str, dict], campaign: str, server="cloud", verbose=False, debug=False) -> None:
    """
    Train a campaign remotely using twinLab
    params:
        filepath_or_params: str or dict; filepath to json or parameters dict for training
        campaign: str; name of this campaign and final trained model (user choice)
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_train_campaign_url(server)
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Campaign"] = campaign
    if isinstance(filepath_or_params, str):
        filepath = filepath_or_params
        with open(filepath) as f:
            params = json.load(f)
    else:
        params = filepath_or_params
    params = utils.coerce_params_dict(params)
    r = requests.post(url, json=params, headers=headers)
    utils.check_response(r)
    if verbose:
        utils.print_response_message(r)


def query_campaign(campaign_name: str, server="cloud", verbose=False, debug=False) -> dict:
    """
    Print summary statistics for a pre-trained campaign
    params:
        campaign_name: str; name of trained model to query
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/query_campaign"
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Campaign"] = campaign_name
    r = requests.get(url, headers=headers)
    utils.check_response(r)
    metadata = utils.extract_item_from_response(r, "metadata")
    if verbose:
        utils.print_response_message(r)
        print("Metadata:")
        pprint(metadata, compact=True, sort_dicts=False)
    return metadata


def list_campaigns(server="cloud", verbose=False, debug=False) -> Union[list, None]:
    """
    List all trained campaigns stored in cloud
    params:
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/list_campaigns"
    headers = utils.construct_standard_headers(debug=debug)
    r = requests.get(url, headers=headers)
    utils.check_response(r)
    if verbose:
        utils.print_response_message(r)
    campaigns = r.json()["campaigns"]
    if verbose and campaigns:
        print("Campaigns:")
        pprint(campaigns)
    return campaigns


def predict_campaign(
    filepath_or_df, campaign: str, server="cloud", verbose=False, debug=False
) -> tuple:
    """
    Predict from a pre-trained campaign that exists on the cloud
    params:
        filepath: str; location of csv dataset on local machine for evaluation
        campaign: str; name of pre-trained model to do the evaluating
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(
        server) + "/predict_campaign"
    if isinstance(filepath_or_df, pd.DataFrame):  # Data frames
        buffer = io.BytesIO()
        filepath_or_df.to_csv(buffer, index=False)
        buffer = buffer.getvalue()
        files = {"file": ("tmp.csv", buffer, "text/csv")}
    else:  # File paths
        files = {"file": (filepath_or_df, open(
            filepath_or_df, "rb"), "text/csv")}
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Campaign"] = campaign
    r = requests.post(url, files=files, headers=headers)
    utils.check_response(r)
    df_mean = utils.extract_csv_from_response(r, "y_mean")
    df_std = utils.extract_csv_from_response(r, "y_std")
    if verbose:
        utils.print_response_message(r)
        print("Mean: \n", df_mean, "\n")
        print("Std: \n", df_std, "\n")
    return df_mean, df_std


def delete_campaign(campaign_name: str, server="cloud", verbose=False, debug=False) -> None:
    """
    Delete campaign directory from S3
    params:
        campaign_name: str; name of trained model to delete
        server: str; either "cloud" or "local"
        verbose: bool
        debug: bool
    """
    url = utils.get_server_url(server) + "/delete_campaign"
    headers = utils.construct_standard_headers(debug=debug)
    headers["X-Campaign"] = campaign_name
    r = requests.post(url, headers=headers)
    utils.check_response(r)
    if verbose:
        utils.print_response_message(r)
